------------------------------------------------------------------------------
--- Library to support CGI programming in the HTML library.
--- It is only intended as an auxiliary library to implement dynamic web
--- pages according to the HTML library.
--- It contains a simple script that is installed for a dynamic
--- web page and which sends the user input to the real application
--- server implementing the application.
---
--- The installation of a cgi script written with this library
--- can be done by the commands
--- <pre>
---   pakcs -m 'startCgiScript "<key>" "<serverprog>"' -s HtmlCgi"
---   mv HtmlCgi.state <cgifile>
--- </pre>
--- where <cgifile> is the desired location of the compiled cgi script,
--- <key> is some key to identify the compilation of this script
--- (together with the application server), and <serverprog> is the name
--- of the application server that might be implicitly started by this
--- script if the server is not active.
--- 
--- @author Michael Hanus
--- @version March 2007
------------------------------------------------------------------------------

module HtmlCgi(CgiServerMsg(..),runCgiServerCmd,cgikey2portname) where

import System
import Read
import Char
import Ports

--------------------------------------------------------------------------
-- Should the output of the server stored in a log file?
logServer = True

--------------------------------------------------------------------------
--- The messages to comunicate between the cgi script and the server program.
--- CgiSubmit env cgienv nextpage - pass the environment and show next page,
---   where env are the values of the environment variables of the web script
---   (e.g., QUERY_STRING, REMOTE_HOST, REMOTE_ADDR),
---   cgienv are the values in the current form submitted by the client,
---   and nextpage is the answer text to be shown in the next web page
--- @cons SketchStatus - get a sketch of the status of the server
--- @cons ShowStatus - show the status of the server with all event handlers
--- @cons CleanServer - clean up the server (with possible termination)
--- @cons StopCgiServer - stop the server
data CgiServerMsg = CgiSubmit [(String,String)] [(String,String)] String
                  | ShowStatus String
                  | SketchStatus String
                  | CleanServer
                  | StopCgiServer

--------------------------------------------------------------------------
-- Read arguments and start script:
main = do
  args <- getArgs
  case args of
    [cgikey,serverprog] -> startCgiScript cgikey (Just serverprog)
    [portname] -> cgiScript portname Nothing  -- for interactive execution
    _ -> putStrLn $ "ERROR: cgi script called with illegal arguments!"

startCgiScript cgikey serverprog  =
  cgiScript (cgikey2portname cgikey ++"@localhost") serverprog

runCgiServerCmd portname cmd
 | cmd=="stop" = do
    putStrLn $ "Trying to stop server at port " ++ portname ++ "..."
    sendScriptServerMessage portname Nothing StopCgiServer
 | cmd=="clean" = do
    putStrLn $ "Trying to clean server at port " ++ portname ++ "..."
    sendScriptServerMessage portname Nothing CleanServer
 | cmd=="show" = do
    sendScriptServerMessage portname Nothing (ShowStatus answer)
    putStrLn answer
 | cmd=="sketch" = do
    sendScriptServerMessage portname Nothing (SketchStatus answer)
    putStrLn answer
 | otherwise = putStrLn $ "ERROR: cgi script called with illegal command!"
 where
  answer free

--- Translates a cgi progname and key into a name for a port:
cgikey2portname cgikey =
  concatMap (\c->if isAlphaNum c then [c] else []) cgikey
  --map (\c->if isAlphaNum c then c else '_') (cgiprog ++ "_" ++cgikey)

-- Simple script to forward user inputs:
cgiScript :: String -> Maybe String -> IO ()
cgiScript portname serverprog = do
  cgiEnvVals <- mapIO getEnviron cgiEnvVars
  newcenv    <- getFormVariables
  sendScriptServerMessage portname serverprog
           (CgiSubmit (zip cgiEnvVars cgiEnvVals) newcenv answer)
  putStrLn answer
 where
  answer free

--- The values of the environment variables of the web script server
--- that are transmitted to the application program.
--- Currently, it contains only a selection of all reasonable variables
--- but this list can be easily extended.
cgiEnvVars = ["QUERY_STRING","REMOTE_HOST","REMOTE_ADDR"]

-- The timeout (in msec) of the script server.
-- If the application server does not answer within the timeout period,
-- we assume that the application server does not exist and we start a new one.
-- After the start, we wait for scriptServerTimeOut2 milliseconds, before
-- we finally give up.
scriptServerTimeOut = 500
scriptServerTimeOut2 = 2000

sendScriptServerMessage portname Nothing msg =
  connectPortRepeat scriptServerTimeOut done 1 portname >>=
  maybe done (\port -> doSend msg port)
sendScriptServerMessage portname (Just serverprog) msg =
  connectPortRepeat scriptServerTimeOut done 1 portname >>=
  maybe (startScriptServer serverprog >>
         connectPortRepeat scriptServerTimeOut2 done 1 portname >>=
         maybe (putStrLn "ERROR: cannot start cgi script server")
               (\port -> doSend msg port))
        (\port -> doSend msg port)

-- Start the server for this script:
startScriptServer prog = do
  if logServer
   then system (prog++" > "++prog++".log 2>&1 &")
   else system (prog++" > /dev/null 2>&1 &")
  --sleep 1

------------------------------------------------------------------------------
--- Gets the list of variable/value pairs sent from the browser for the
--- current CGI script.
--- Used for the implementation of the HTML event handlers.
getFormVariables :: IO [(String,String)]
getFormVariables = do
  clen <- getEnviron "CONTENT_LENGTH"
  cont <- getNChar (readNat clen)
  return (includeCoordinates (parseCgiEnv cont))

-- translate a string of cgi environment bindings into list of binding pairs:
parseCgiEnv :: String -> [(String,String)]
parseCgiEnv s | s == ""   = []
              | otherwise = map ufield2field
                             (map (\(n,v)->(n,urlencoded2string v))
                                  (map (splitChar '=') (split (=='&') s)))
 where
   ufield2field (n,v) = if take 7 n == "UFIELD_"
                        then (tail n, urlencoded2string v)
                        else (n,v)

   -- split a string at particular character:
   splitChar c xs = let (ys,zs) = break (==c) xs
                    in if zs==[] then (ys,zs) else (ys,tail zs)

   -- split a string at all positions of a particular character:
   split p xs =
    let (ys,zs) = break p xs
    in if zs==[] then [ys]
                 else ys : split p (tail zs)

--- Translates urlencoded string into equivalent ASCII string.
urlencoded2string :: String -> String
urlencoded2string [] = []
urlencoded2string (c:cs)
  | c == '+'  = ' ' : urlencoded2string cs
  | c == '%'  = chr (readHex (take 2 cs)) : urlencoded2string (drop 2 cs)
  | otherwise = c : urlencoded2string cs

includeCoordinates :: [(String,String)] -> [(String,String)]
includeCoordinates [] = []
includeCoordinates ((tag,val):cenv) 
  = case break (=='.') tag of
      (_,[]) -> (tag,val):includeCoordinates cenv
      (event,['.','x']) -> ("x",val):(event,val):includeCoordinates cenv
      (_,['.','y']) -> ("y",val):includeCoordinates cenv
      _ -> error "includeCoordinates: unexpected . in url parameter"
   

-- get n chars from stdin:
getNChar n = if n<=0 then return ""
                     else do c <- getChar
                             cs <- getNChar (n-1)
                             return (c:cs)

------------------------------------------------------------------------------
